# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'thread' # for Mutex
require 'singleton'
require 'sqreen/ecosystem/loggable'

# see https://github.com/sqreen/TechDoc/blob/master/content/specs/spec000024-sampling.md

module Sqreen
  module Ecosystem
    module Tracing
      class Sampler
        # @param [Array<Hash{String=>Object}>] definition
        def initialize(definition)
          @lines = definition.map { |h| Line.new(h) }
        end

        def should_sample?
          line = @lines.find(&:triggers?)
          line ? line.saved_definition : false
        end

        class Line
          include Loggable

          attr_reader :saved_definition

          # @param [Hash{String=>Object}] definition
          def initialize(definition)
            @saved_definition = definition
            @primitives = []

            unknown = definition.keys - PRIMITIVES_MAP.keys
            unless unknown.empty?
              logger.warn "Unknown primitives: #{unknown}"
              @primitives << AlwaysFalsePrimitive.instance
              return
            end

            PRIMITIVES_MAP.each do |key, prim_class|
              next unless definition[key]
              @primitives << prim_class.new(definition[key])
            end
            # if @primitives is empty the line will always
            # return true: [].all?(&:triggers?) is true
          end

          def triggers?
            @primitives.all?(&:triggers?)
          end
        end

        class AlwaysFalsePrimitive
          include Singleton

          def triggers?
            false
          end
        end

        class CallsPrimitive
          def initialize(calls_period)
            @calls_period = calls_period
            @count = 0
            @mutex = Mutex.new
          end

          def triggers?
            prev_count = nil
            @mutex.synchronize do
              prev_count = @count
              @count += 1
            end

            (prev_count % @calls_period).zero?
          end
        end

        class RandomPrimitive
          def initialize(probability)
            @probability = probability
          end

          def triggers?
            @probability >= rand
          end
        end

        class MaxDurationMinutesPrimitive
          def initialize(time_in_minutes)
            @deadline = Sqreen.time + time_in_minutes * 60
            @passed = false # no locking needed
          end

          def triggers?
            return false if @passed
            if Sqreen.time > @deadline
              @passed = true
              return false
            end
            true
          end
        end

        class TargetPerMinutePrimitive
          def initialize(max_calls)
            @max_calls = max_calls
            @minute_last_call = cur_minute
            @calls_accumulated = 0
            @mutex = Mutex.new
          end

          def triggers?
            this_minute = cur_minute
            calls_cur_minute = @mutex.synchronize do
              if @minute_last_call == this_minute
                @calls_accumulated += 1
              else
                @minute_last_call = this_minute
                @calls_accumulated = 1
              end
            end

            calls_cur_minute <= @max_calls
          end

          private

          def cur_minute
            (Sqreen.time / 60).floor
          end
        end

        class MaxCallsPrimitive
          def initialize(max_calls)
            @max_calls = max_calls
            @disabled = false # to avoid lock
            @mutex = Mutex.new
            @num_calls = 0
          end

          def triggers?
            return false if @disabled
            num_calls = @mutex.synchronize do
              @num_calls += 1
            end

            num_calls <= @max_calls
          end
        end

        PRIMITIVES_MAP = {
          "calls"                => CallsPrimitive,
          "random"               => RandomPrimitive,
          "max_duration_minutes" => MaxDurationMinutesPrimitive,
          "target_per_minute"    => TargetPerMinutePrimitive,
          "max_calls"            => MaxCallsPrimitive,
        }.freeze
      end
    end
  end
end
