# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'thread'
require 'sqreen/ecosystem/loggable'
require 'sqreen/ecosystem/tracing/sampler'

module Sqreen
  module Ecosystem
    module Tracing
      # tracing sampling configuration, as specified by the 2nd argument of the
      # tracing_enable command.
      # See https://github.com/sqreen/TechDoc/blob/master/content/specs/spec000025-enabling-tracing.md
      class SamplingConfiguration
        include Loggable

        DEFAULT_SCOPE = '*'.freeze

        def initialize(sampling_config)
          @sampling_config = sampling_config
          @samplers = {}
          @samplers_virtual = build_virtual_holders(sampling_config)
        end

        # @param [String] scope either the scope or the part behind @ in a virtual scope
        # @param [String] qualifier the part after @ in a virtual scope or nil
        def should_sample?(scope, qualifier = nil)
          if qualifier
            fetch_sampler_virtual(scope, qualifier).should_sample?
          else
            fetch_sampler(scope).should_sample?
          end
        end

        def forget_virtual_scope(scope, qualifier)
          holder = @samplers_virtual[scope]
          return unless holder.delete!(qualifier)
        end

        private

        def fetch_sampler_virtual(scope, qualifier)
          holder = @samplers_virtual[scope]

          # no virtual scope configured, fallback to plain scope
          return fetch_sampler(scope) unless holder

          holder[qualifier]
        end

        def fetch_sampler(scope)
          sampler = @samplers[scope]
          return sampler if sampler

          cfg = @sampling_config[scope] || @sampling_config[DEFAULT_SCOPE]
          if cfg
            @samplers[scope] = SamplingConfiguration.build_sampler(scope, cfg)
          else
            logger.info { "Disabling scope #{scope} due to its not being configured" }
            @samplers[scope] = DisabledScopeSampler
          end
        end

        def build_virtual_holders(sampling_config)
          Hash[
            sampling_config
            .select { |scope, _cfg| scope.end_with?('@*') }
            .map do |scope, cfg|
              parent = scope[0...-2] # remove trailing @*
              [parent, VirtualScopesHolder.new(parent, cfg)]
            end
          ]
        end

        class << self
          include Loggable

          def build_sampler(scope, cfg)
            do_build_sampler(cfg['enabled'], cfg['sampling'] || [{}])
          rescue StandardError => e
            logger.warn "Invalid sampling configuration for #{scope}: #{e.inspect}"
            logger.debug { e.backtrace }
            DisabledScopeSampler
          end

          private

          def do_build_sampler(enabled, sampling)
            if enabled
              Sampler.new(sampling)
            else
              logger.debug { "Disabling scope #{scope}" }
              DisabledScopeSampler
            end
          end
        end
      end

      class VirtualScopesHolder
        include Loggable

        MAX_VIRTUAL_SCOPES = 120
        DISCARD_SIZE = 20

        def initialize(parent, cfg)
          @parent = parent
          @cfg = cfg
          @virtual_scopes = {}
          @mutex = Mutex.new
        end

        def [](qualifier)
          return false unless @cfg['enabled']

          @mutex.synchronize do
            sampler = @virtual_scopes[qualifier]
            return sampler if sampler

            @virtual_scopes[qualifier] = create_virtual(qualifier)
          end
        end

        def delete!(qualifier)
          @virtual_scopes.delete(qualifier)
        end

        private

        def create_virtual(qualifier)
          discard if @virtual_scopes.size >= MAX_VIRTUAL_SCOPES

          @virtual_scopes[qualifier] =
            SamplingConfiguration.build_sampler("#{@parent}@#{qualifier}", @cfg)
        end

        def discard
          logger.info { "Discarding excess virtual scopes for scope '#{@parent}'" }
          discard_keys = @virtual_scopes.keys.sample(DISCARD_SIZE)
          @virtual_scopes.delete_if { |k, _v| discard_keys.include? k }
        end
      end

      # fake sampler that always returns false
      class DisabledScopeSampler
        class << self
          def should_sample?
            false
          end
        end
      end
    end
  end
end
