# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/log/loggable'
require 'sqreen/ecosystem'
require 'sqreen/ecosystem/dispatch_table'
require 'sqreen/ecosystem_integration/instrumentation_service'
require 'sqreen/ecosystem_integration/request_lifecycle_tracking'
require 'sqreen/ecosystem_integration/signal_consumption'

module Sqreen
  # This class is the interface through which the agent interacts
  # with the ecosystem.
  #
  # Other classes in the EcosystemIntegration module implement the
  # functionality that the ecosystem requires in order to deliver
  # data to the agent and to be informed by the agent of certain
  # key events (see Sqreen::Ecosystem::DispatchTable).
  class EcosystemIntegration
    include Sqreen::Log::Loggable

    # @param [Sqreen::Framework] framework
    def initialize(framework, queue)
      @framework = framework
      @queue = queue
      @request_lifecycle = RequestLifecycleTracking.new
      @online = false
    end

    def init
      raise 'already initialized' if @online

      setup_dispatch_table
      Ecosystem.init
      logger.info 'Ecosystem successfully initialized'
      @online = true
    rescue ::Exception => e # rubocop:disable Lint/RescueException
      logger.warn { "Error initializing Ecosystem: #{e.message}" }
      logger.debug { e.backtrace.map { |x| "  #{x}" }.join("\n") }
      Sqreen::RemoteException.record(e)
    end

    def disable
      raise NotImplementedYet
    end

    def request_start(rack_request)
      return unless @online

      Ecosystem.start_transaction
      @request_lifecycle.notify_request_start(rack_request)
    end

    def request_end
      return unless @online

      Ecosystem.end_transaction
    end

    def handle_tracing_command(trace_id_prefix, scopes_config)
      return unless @online

      Ecosystem.configure_sampling(trace_id_prefix, scopes_config)
    end

    private

    def setup_dispatch_table
      Ecosystem::DispatchTable.consume_signal =
        create_signal_consumption.method(:consume_signal)

      Ecosystem::DispatchTable.add_request_start_listener =
        @request_lifecycle.method(:add_start_observer)

      Ecosystem::DispatchTable.fetch_logger = lambda { logger }

      Ecosystem::DispatchTable.instrument = InstrumentationService.method(:instrument)
    end

    def create_signal_consumption
      SignalConsumption.new(@framework, @request_lifecycle, @queue)
    end
  end
end
