# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/backport/clock_gettime'
require 'sqreen/graft'

module Sqreen
  module Graft
    CallbackCall = Struct.new(:callback, :remaining, :instance, :args, :raised, :returned)
    HookedCall = Struct.new(:instance, :args_passed, :args_pass, :raised, :returned, :return, :returning, :retrying, :args_passing, :raising, :raise)

    class Ball
      def return(value)
        Flow.return(value)
      end

      def args(value)
        Flow.args(value)
      end

      def retry
        Flow.retry
      end

      def raise(value)
        Flow.raise(value)
      end
    end

    class Flow
      class << self
        def return(value)
          new(:return, value)
        end

        def args(value)
          new(:args, value)
        end

        def retry
          new(:retry)
        end

        def raise(value)
          new(:raise, value)
        end
      end

      def initialize(action, value, brk = false)
        @action = action
        @value = value
        @break = brk
      end

      def return?
        @action == :return
      end

      def return
        @value if return?
      end

      def raise?
        @action == :raise
      end

      def raise
        @value if raise?
      end

      def args?
        @action == :args
      end

      def args
        @value if args?
      end

      def retry?
        @action == :retry
      end

      def break!
        @break = true

        self
      end

      def break?
        @break ? true : false
      end
    end

    class TimerError < StandardError; end

    class Timer
      def self.read
        Process.clock_gettime(Process::CLOCK_MONOTONIC)
      end

      attr_reader :tag, :size

      def initialize(tag, &block)
        @tag = tag
        @block = block
        @tally = 0
        @size = 0
      end

      def elapsed
        raise(TimerError, 'Timer#elapsed when paused') if @size.even?

        @tally + Timer.read
      end

      def duration
        raise(TimerError, 'Timer#duration when running') if @size.odd?

        @tally
      end

      def ignore
        raise(TimerError, 'Timer#ignore when paused') if @size.even?

        @size += 1
        @tally += Timer.read
        yield(self)
      ensure
        @size += 1
        @tally -= Timer.read
      end

      def measure(opts = nil)
        raise(TimerError, 'Timer#measure when running') if @size.odd?

        now = Timer.read

        ignore = opts[:ignore] if opts
        if ignore
          ignore.size += 1
          ignore.tally += now
        end

        @size += 1
        @tally -= now

        yield(self)
      ensure
        now = Timer.read

        if ignore
          ignore.size += 1
          ignore.tally -= now
        end

        @size += 1
        @tally += now

        @block.call(self) if @block
      end

      def start(at = Timer.read)
        raise(TimerError, 'Timer#start when started') unless @size.even?

        @size += 1
        @tally -= at

        at
      end

      def stop(at = Timer.read)
        raise(TimerError, 'Timer#stop when unstarted') unless @size.odd?

        @size += 1
        @tally += at

        at
      end

      def started?
        @size != 0 && @size.odd?
      end

      def stopped?
        @size != 0 && @size.even?
      end

      def running?
        @size.odd?
      end

      def paused?
        @size.even?
      end

      def include_measurements(another_timer)
        @blips += another_timer.instance_variable_get(:@blips)
      end

      def start_and_end
        raise 'Not exactly two measurements recorded' unless size == 2
        @blips
      end

      def to_s
        "#{@tag}: time=%.03fus" % (duration * 1_000_000)
      end

      protected

      attr_reader :tally
      attr_writer :size, :tally
    end
  end
end
