# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/graft/call'

module Sqreen
  module Graft
    class Hook
      def self.wrapper(hook)
        timed_hooks_proc = proc do |t|
          if (request = Thread.current[:sqreen_http_request])
            request[:timed_hooks] << t if request[:timed_level] >= 1
          end
        end
        timed_callbacks_proc = proc do |t|
          if (request = Thread.current[:sqreen_http_request])
            request[:timed_callbacks] << t if request[:timed_level] >= 1
          end
        end

        Proc.new do |*args, **kwargs, &block|
          request = Thread.current[:sqreen_http_request]

          if Thread.current[:sqreen_hook_entered]
            if hook.point.super?
              return super(*args, **kwargs, &block)
            else
              return hook.point.apply(self, 'sqreen_hook', *args, **kwargs, &block)
            end
          end

          if request && request[:time_budget_expended] && (hook.before + hook.after + hook.raised + hook.ensured).none?(&:mandatory)
            if request[:timed_level] >= 2
              begin
                request[:skipped_callbacks].concat(hook.before)

                if hook.point.super?
                  return super(*args, **kwargs, &block)
                else
                  return hook.point.apply(self, 'sqreen_hook', *args, **kwargs, &block)
                end
              rescue ::Exception # rubocop:disable Lint/RescueException
                request[:skipped_callbacks].concat(hook.raised)
                raise
              else
                request[:skipped_callbacks].concat(hook.after)
              ensure
                request[:skipped_callbacks].concat(hook.ensured)
              end
            else
              if hook.point.super? # rubocop:disable Style/IfInsideElse
                return super(*args, **kwargs, &block)
              else
                return hook.point.apply(self, 'sqreen_hook', *args, **kwargs, &block)
              end
            end
          end

          hook_point_super = hook.point.super?
          logger = Sqreen::Graft.logger
          logger_debug = Sqreen::Graft.logger.debug?

          Timer.new(hook.point, &timed_hooks_proc).measure do |chrono|
            # budget implies request
            # TODO: make budget depend on a generic context (currently "request")
            budget = request[:time_budget] if request
            if budget
              budget_threshold = request[:time_budget_threshold]
              budget_ratio = request[:time_budget_ratio]
              sqreen_timer = request[:sqreen_timer]
              request_timer = request[:request_timer]
            end

            hooked_call = HookedCall.new(self, args)

            begin
              begin
                sqreen_timer.start if budget
                Thread.current[:sqreen_hook_entered] = true

                # TODO: make Call have #ball to throw by cb
                # TODO: can Call be the ball? r = catch(Call.new, &c)
                # TODO: is catch return value a Call? a #dispatch?
                # TODO: make before/after/raised return a CallbackCollection << Array (or extend with module)
                # TODO: add CallbackCollection#each_with_call(instance, args) { |call| ... } ?
                # TODO: HookCall x CallbackCollection#each_with_call x Flow
                # TODO: TimedHookCall TimedCallbackCall
                # TODO: TimeBoundHookCall TimeBoundCallbackCall TimeBoundFlow?

                request_elapsed = request_timer.elapsed if budget

                hook.before.each do |c|
                  next if c.ignore && c.ignore.call

                  if budget && !c.mandatory
                    sqreen_elapsed = sqreen_timer.elapsed
                    if budget_ratio && !request[:time_budget_expended]
                      fixed_budget = budget_threshold
                      proportional_budget = (request_elapsed - sqreen_elapsed) * budget_ratio
                      remaining = (fixed_budget > proportional_budget ? fixed_budget : proportional_budget) - sqreen_elapsed
                    else
                      remaining = budget_threshold - sqreen_elapsed
                    end
                    unless remaining > 0
                      request[:skipped_callbacks] << c
                      request[:time_budget_expended] = true
                      next
                    end
                  end

                  flow = catch(Ball.new) do |ball|
                    Timer.new(c.name, &timed_callbacks_proc).measure(ignore: chrono) do
                      c.call(CallbackCall.new(c, remaining, hooked_call.instance, hooked_call.args_passed), ball)
                    end
                  end

                  next unless c.flow && flow.is_a?(Flow)
                  hooked_call.raise = flow.raise and hooked_call.raising = true if flow.raise?
                  hooked_call.args_pass = flow.args and hooked_call.args_passing = true if flow.args?
                  hooked_call.return = flow.return and hooked_call.returning = true if flow.return?
                  break if flow.break?
                end unless hook.disabled?
              rescue StandardError => e
                Sqreen::Weave.logger.debug { "exception:#{e.class} when:before message:'#{e.message}' location:\"#{e.backtrace[0]}\"" }
                Sqreen::RemoteException.record(e) if Sqreen.queue
              end

              raise hooked_call.raise if hooked_call.raising
              return hooked_call.return if hooked_call.returning
            ensure
              Thread.current[:sqreen_hook_entered] = false
              sqreen_timer.stop if budget
            end unless hook.before.empty?

            begin
              chrono.ignore do
                if hook_point_super
                  hooked_call.returned = super(*(hooked_call.args_passing ? hooked_call.args_pass : hooked_call.args_passed), **kwargs, &block)
                else
                  hooked_call.returned = hook.point.apply(hooked_call.instance, 'sqreen_hook', *(hooked_call.args_passing ? hooked_call.args_pass : hooked_call.args_passed), **kwargs, &block)
                end
              end
            rescue ::Exception => e # rubocop:disable Lint/RescueException
              begin
                sqreen_timer.start if budget
                Thread.current[:sqreen_hook_entered] = true
                hooked_call.raised = e

                logger.debug { "[#{Process.pid}] Hook #{hook.point} disabled:#{hook.disabled?} exception:#{e}" } if logger_debug

                # TODO: early escape causes raise too early then caught by ensure which mistakenly reports legit errors to sqreen
                # raise if hook.raised.empty?

                request_elapsed = request_timer.elapsed if budget

                hook.raised.each do |c|
                  next if c.ignore && c.ignore.call

                  if budget && !c.mandatory
                    sqreen_elapsed = sqreen_timer.elapsed
                    if budget_ratio && !request[:time_budget_expended]
                      fixed_budget = budget_threshold
                      proportional_budget = (request_elapsed - sqreen_elapsed) * budget_ratio
                      remaining = (fixed_budget > proportional_budget ? fixed_budget : proportional_budget) - sqreen_elapsed
                    else
                      remaining = budget_threshold - sqreen_elapsed
                    end
                    unless remaining > 0
                      request[:skipped_callbacks] << c
                      request[:time_budget_expended] = true
                      next
                    end
                  end

                  flow = catch(Ball.new) do |ball|
                    Timer.new(c.name, &timed_callbacks_proc).measure(ignore: chrono) do
                      c.call(CallbackCall.new(c, remaining, hooked_call.instance, hooked_call.args_passing ? hooked_call.args_pass : hooked_call.args_passed, hooked_call.raised), ball)
                    end
                  end

                  next unless c.flow && flow.is_a?(Flow)
                  hooked_call.raise = flow.raise and hooked_call.raising = true if flow.raise?
                  hooked_call.return = flow.return and hooked_call.returning = true if flow.return?
                  hooked_call.retrying = true if flow.retry?
                  break if flow.break?
                end unless hook.disabled?
              rescue StandardError => e
                Sqreen::Weave.logger.debug { "exception:#{e.class} when:raised message:'#{e.message}' location:\"#{e.backtrace[0]}\"" }
                Sqreen::RemoteException.record(e) if Sqreen.queue
              end

              retry if hooked_call.retrying
              raise hooked_call.raise if hooked_call.raising
              return hooked_call.return if hooked_call.returning
              raise
            else
              begin
                sqreen_timer.start if budget
                Thread.current[:sqreen_hook_entered] = true

                # TODO: hooked_call.returning should be always false here?
                return hooked_call.returning ? hooked_call.return : hooked_call.returned if hook.after.empty?

                request_elapsed = request_timer.elapsed if budget

                hook.after.each do |c|
                  next if c.ignore && c.ignore.call

                  if budget && !c.mandatory
                    sqreen_elapsed = sqreen_timer.elapsed
                    if budget_ratio && !request[:time_budget_expended]
                      fixed_budget = budget_threshold
                      proportional_budget = (request_elapsed - sqreen_elapsed) * budget_ratio
                      remaining = (fixed_budget > proportional_budget ? fixed_budget : proportional_budget) - sqreen_elapsed
                    else
                      remaining = budget_threshold - sqreen_elapsed
                    end
                    unless remaining > 0
                      request[:skipped_callbacks] << c
                      request[:time_budget_expended] = true
                      next
                    end
                  end

                  flow = catch(Ball.new) do |ball|
                    Timer.new(c.name, &timed_callbacks_proc).measure(ignore: chrono) do
                      c.call(CallbackCall.new(c, remaining, hooked_call.instance, hooked_call.args_passing ? hooked_call.args_pass : hooked_call.args_passed, nil, hooked_call.returned), ball)
                    end
                  end

                  next unless c.flow && flow.is_a?(Flow)
                  hooked_call.raise = flow.raise and hooked_call.raising = true if flow.raise?
                  hooked_call.return = flow.return and hooked_call.returning = true if flow.return?
                  break if flow.break?
                end unless hook.disabled?
              rescue StandardError => e
                Sqreen::Weave.logger.debug { "exception:#{e.class} when:after message:'#{e.message}' location:\"#{e.backtrace[0]}\"" }
                Sqreen::RemoteException.record(e) if Sqreen.queue
              end

              raise hooked_call.raise if hooked_call.raising
              return hooked_call.returning ? hooked_call.return : hooked_call.returned
            ensure
              begin
                # TODO: sqreen_timer.start if someone has thrown?
                # TODO: sqreen_timer.stop at end of rescue+else?
                # TODO: Thread.current[:sqreen_hook_entered] = true if neither rescue nor else ie thrown?
                # TODO: Thread.current[:sqreen_hook_entered] = false at end of rescue+else? (risky?)

                # TODO: uniform early bail out? (but don't forget sqreen_hook_entered and sqreen_timer)
                # return hooked_call.returning ? hooked_call.return : hooked_call.returned if hook.ensured.empty?

                # done at either rescue or else
                # request_elapsed = request_timer.elapsed if budget # && !hook.ensured.empty?

                hook.ensured.each do |c|
                  next if c.ignore && c.ignore.call

                  if budget && !c.mandatory
                    sqreen_elapsed = sqreen_timer.elapsed
                    if budget_ratio && !request[:time_budget_expended]
                      fixed_budget = budget_threshold
                      proportional_budget = (request_elapsed - sqreen_elapsed) * budget_ratio
                      remaining = (fixed_budget > proportional_budget ? fixed_budget : proportional_budget) - sqreen_elapsed
                    else
                      remaining = budget_threshold - sqreen_elapsed
                    end
                    unless remaining > 0
                      request[:skipped_callbacks] << c
                      request[:time_budget_expended] = true
                      next
                    end
                  end

                  flow = catch(Ball.new) do |ball|
                    Timer.new(c.name, &timed_callbacks_proc).measure(ignore: chrono) do
                      c.call(CallbackCall.new(c, remaining, hooked_call.instance, hooked_call.args_passing ? hooked_call.args_pass : hooked_call.args_passed, nil, hooked_call.returned), ball)
                    end
                  end

                  next unless c.flow && flow.is_a?(Flow)
                  hooked_call.raise = flow.raise and hooked_call.raising = true if flow.raise?
                  hooked_call.return = flow.return and hooked_call.returning = true if flow.return?
                  break if flow.break?
                end unless hook.ensured.empty? || hook.disabled?

                Thread.current[:sqreen_hook_entered] = false
                sqreen_timer.stop if budget
              rescue StandardError => e
                Sqreen::Weave.logger.debug { "exception:#{e.class} when:ensured message:'#{e.message}' location:\"#{e.backtrace[0]}\"" }
                Sqreen::RemoteException.record(e) if Sqreen.queue
              end

              # TODO: should we run the following?
              # raise hooked_call.raise if hooked_call.raising
              # return hooked_call.returning ? hooked_call.return : hooked_call.returned
            end
          end # chrono
        end
      end
    end
  end
end
