# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/dependency'
require 'sqreen/backport/original_name'
require 'sqreen/graft'
require 'sqreen/graft/hook_point_error'

module Sqreen
  module Graft
    class HookSpot < Module
      def initialize(key)
        @key = key
      end

      attr_reader :key

      def inspect
        "#<#{self.class.name}: #{@key.inspect}>"
      end
    end

    class HookPoint
      DEFAULT_STRATEGY = Module.respond_to?(:prepend) ? :prepend : :chain

      def self.parse(hook_point)
        klass_name, separator, method_name = hook_point.split(/(\#|\.)/, 2)

        raise ArgumentError, hook_point if klass_name.nil? || separator.nil? || method_name.nil?
        raise ArgumentError, hook_point unless ['.', '#'].include?(separator)

        method_kind = separator == '.' ? :klass_method : :instance_method

        [klass_name.to_sym, method_kind, method_name.to_sym]
      end

      attr_reader :klass_name, :method_kind, :method_name

      def initialize(hook_point, strategy = DEFAULT_STRATEGY)
        @klass_name, @method_kind, @method_name = Sqreen::Graft::HookPoint.parse(hook_point)
        @strategy = strategy
      end

      def to_s
        @to_s ||= "#{@klass_name}#{@method_kind == :instance_method ? '#' : '.'}#{@method_name}"
      end

      def exist?
        return false unless Sqreen::Dependency.const_exist?(@klass_name)

        if klass_method?
          (klass.singleton_class.public_instance_methods(false) + klass.singleton_class.protected_instance_methods(false) + klass.singleton_class.private_instance_methods(false)).include?(@method_name)
        elsif instance_method?
          (klass.public_instance_methods(false) + klass.protected_instance_methods(false) + klass.private_instance_methods(false)).include?(@method_name)
        else
          Sqreen::Graft.logger.error { "[#{Process.pid}] #{self} unknown hook point kind" }
          raise HookPointError, "#{self} unknown hook point kind"
        end
      end

      def klass
        Sqreen::Dependency.resolve_const(@klass_name)
      end

      def klass_method?
        @method_kind == :klass_method
      end

      def private_method?
        if klass_method?
          klass.private_methods.include?(@method_name)
        elsif instance_method?
          klass.private_instance_methods.include?(@method_name)
        else
          Sqreen::Graft.logger.error { "[#{Process.pid}] #{self} unknown hook point kind" }
          raise HookPointError, "#{self} unknown hook point kind"
        end
      end

      def protected_method?
        if klass_method?
          klass.protected_methods.include?(@method_name)
        elsif instance_method?
          klass.protected_instance_methods.include?(@method_name)
        else
          Sqreen::Graft.logger.error { "[#{Process.pid}] #{self} unknown hook point kind" }
          raise HookPointError, "#{self} unknown hook point kind"
        end
      end

      def instance_method?
        @method_kind == :instance_method
      end

      def installed?(key)
        case @strategy
        when :chain then defined(key)
        when :prepend then prepended?(key) && overridden?(key)
        end
      end

      def super?
        @strategy == :prepend
      end

      def install(key, &block)
        if installed?(key)
          Sqreen::Graft.logger.debug { "[#{Process.pid}] #{self} already installed" }
          return
        end
        unless exist?
          Sqreen::Graft.logger.debug { "[#{Process.pid}] #{self} hook point not found" }
          return
        end

        case @strategy
        when :chain
          define(key, &block)
          chain(key)
        when :prepend
          prepend(key)
          override(key, &block)
        end
      end

      def uninstall(key)
        unless installed?(key)
          Sqreen::Graft.logger.debug { "[#{Process.pid}] #{self} not installed" }
          return
        end
        unless exist?
          Sqreen::Graft.logger.debug { "[#{Process.pid}] #{self} hook point not found" }
          return
        end

        case @strategy
        when :chain
          disable(key)
          remove(key)
        when :prepend
          unoverride(key) if overridden?(key)
        end
      end

      def enable(key)
        case @strategy
        when :chain
          chain(key)
        when :prepend
          raise HookPointError, "enable called on prepend mode"
        end
      end

      def disable(key)
        case @strategy
        when :chain
          unchain(key)
        when :prepend
          unoverride(key)
        end
      end

      def disabled?(key)
        case @strategy
        when :chain
          !chained?(key)
        when :prepend
          !overridden?(key)
        end
      end

      private

      def hook_spot(key)
        target = klass_method? ? klass.singleton_class : klass
        mod = target.ancestors.each { |e| break if e == target; break(e) if e.class == HookSpot && e.key == key }
        raise "Inconsistency detected: #{target} missing from its own ancestors" if mod.is_a?(Array)

        [target, mod]
      end

      def prepend(key)
        target, mod = hook_spot(key)

        mod ||= HookSpot.new(key)

        target.instance_eval { prepend(mod) }
      end

      def prepended?(key)
        _, mod = hook_spot(key)

        mod != nil
      end

      def overridden?(key)
        _, mod = hook_spot(key)

        (mod.instance_methods(false) + mod.protected_instance_methods(false) + mod.private_instance_methods(false)).include?(method_name)
      end

      def override(key, &block)
        hook_point = self
        method_name = @method_name

        _, mod = hook_spot(key)

        mod.instance_eval do
          if hook_point.private_method?
            private
          elsif hook_point.protected_method?
            protected
          else
            public
          end

          define_method(:"#{method_name}", &block)
        end
      end

      def unoverride(key)
        method_name = @method_name

        _, mod = hook_spot(key)

        mod.instance_eval { remove_method(method_name) }
      end

      def defined(suffix)
        if klass_method?
          (klass.methods + klass.protected_methods + klass.private_methods).include?(:"#{method_name}_with_#{suffix}")
        elsif instance_method?
          (klass.instance_methods + klass.protected_instance_methods + klass.private_instance_methods).include?(:"#{method_name}_with_#{suffix}")
        else
          Sqreen::Graft.logger.error { "[#{Process.pid}] #{self} unknown hook point kind" }
          raise HookPointError, "#{self} unknown hook point kind"
        end
      end

      def define(suffix, &block)
        hook_point = self
        method_name = @method_name

        if klass_method?
          klass.singleton_class.instance_eval do
            if hook_point.private_method?
              private
            elsif hook_point.protected_method?
              protected
            else
              public
            end

            define_method(:"#{method_name}_with_#{suffix}", &block)
          end
        elsif instance_method?
          klass.class_eval do
            if hook_point.private_method?
              private
            elsif hook_point.protected_method?
              protected
            else
              public
            end

            define_method(:"#{method_name}_with_#{suffix}", &block)
          end
        else
          raise HookPointError, 'unknown hook point kind'
        end
      end

      def remove(suffix)
        method_name = @method_name

        if klass_method?
          klass.singleton_class.instance_eval do
            remove_method(:"#{method_name}_with_#{suffix}")
          end
        elsif instance_method?
          klass.class_eval do
            remove_method(:"#{method_name}_with_#{suffix}")
          end
        else
          raise HookPointError, 'unknown hook point kind'
        end
      end

      def chained?(suffix)
        method_name = @method_name

        if klass_method?
          klass.singleton_class.instance_eval do
            instance_method(:"#{method_name}").original_name == :"#{method_name}_with_#{suffix}"
          end
        elsif instance_method?
          klass.class_eval do
            instance_method(:"#{method_name}").original_name == :"#{method_name}_with_#{suffix}"
          end
        else
          raise HookPointError, 'unknown hook point kind'
        end
      end

      def chain(suffix)
        method_name = @method_name

        if klass_method?
          klass.singleton_class.instance_eval do
            alias_method :"#{method_name}_without_#{suffix}", :"#{method_name}"
            alias_method :"#{method_name}", :"#{method_name}_with_#{suffix}"
          end
        elsif instance_method?
          klass.class_eval do
            alias_method :"#{method_name}_without_#{suffix}", :"#{method_name}"
            alias_method :"#{method_name}", :"#{method_name}_with_#{suffix}"
          end
        else
          raise HookPointError, 'unknown hook point kind'
        end
      end

      def unchain(suffix)
        method_name = @method_name

        if klass_method?
          klass.singleton_class.instance_eval do
            alias_method :"#{method_name}", :"#{method_name}_without_#{suffix}"
          end
        elsif instance_method?
          klass.class_eval do
            alias_method :"#{method_name}", :"#{method_name}_without_#{suffix}"
          end
        end
      end
    end
  end
end

if RUBY_VERSION =~ /^2\./
  load File.join(__dir__, 'hook_point.ruby_2.rb')
else
  load File.join(__dir__, 'hook_point.ruby_3.rb')
end
