# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/aggregated_metric'
require 'sqreen/log/loggable'
require 'sqreen/legacy/waf_redactions'
require 'sqreen/kit/string_sanitizer'

module Sqreen
  module Legacy
    # see also Sqreen::Signals::SignalsSubmissionStrategy
    # usage in Sqreen:Session
    class OldEventSubmissionStrategy
      include Sqreen::Log::Loggable

      RETRY_MANY = 301

      def initialize(post_proc)
        @post_proc = post_proc
      end

      def post_metrics(metrics)
        return if metrics.nil? || metrics.empty?
        payload = { metrics: metrics.map { |m| EventToHash.convert_agg_metric(m) } }
        post('metrics', payload, {}, RETRY_MANY)
      end

      # @param attack [Sqreen::Attack]
      def post_attack(attack)
        post('attack', EventToHash.convert_attack(attack), {}, RETRY_MANY)
      end

      # @param [Sqreen::RequestRecord] request_record
      def post_request_record(request_record)
        rr_hash = EventToHash.convert_request_record(request_record)
        post('request_record', rr_hash, {}, RETRY_MANY)
      end

      # Post an exception to Sqreen for analysis
      # @param exception [RemoteException] Exception and context to be sent over
      def post_sqreen_exception(exception)
        data = EventToHash.convert_exception(exception)
        post('sqreen_exception', data, {}, 5)
      rescue StandardError => e
        logger.warn(format('Could not post exception (network down? %s) %s',
                           e.inspect,
                           exception.inspect))
        nil
      end

      def post_batch(events)
        batch = events.map do |event|
          h = case event
              when AggregatedMetric
                logger.warn "Aggregated metric event in non-signal mode. Signals disabled at runtime?"
                next
              when Sqreen::Kit::Signals::Signal
                logger.warn "Signal event in non-signal mode"
                next
              when Sqreen::Kit::Signals::Trace
                logger.warn "Trace event in non-signal mode"
                next
              when Attack # in practice only found inside req rec
                EventToHash.convert_attack event
              when RemoteException
                EventToHash.convert_exception event
              when RequestRecord
                EventToHash.convert_request_record event
              else
                logger.warn "Unexpected event type: #{event}"
                next
              end
          h['event_type'] = event_kind(event)
          h
        end
        Sqreen.log.debug do
          tally = Hash[events.group_by(&:class).map { |k, v| [k, v.count] }]
          "Doing batch with the following tally of event types: #{tally}"
        end
        post('batch', { batch: batch.compact }, {}, RETRY_MANY)
      end

      private

      # see +Sqreen::Session.post+
      def post(*args)
        @post_proc[*args]
      end

      def event_kind(event)
        case event
        when Sqreen::RemoteException then 'sqreen_exception'
        when Sqreen::Attack then 'attack'
        when Sqreen::RequestRecord then 'request_record'
        end
      end
    end

    module EventToHash
      class << self
        # @param attack [Sqreen::Attack]
        def convert_attack(attack)
          payload = attack.payload
          res = {}
          rule_p = payload['rule']
          request_p = payload['request']
          res[:rule_name]    = rule_p['name']         if rule_p && rule_p['name']
          res[:rulespack_id] = rule_p['rulespack_id'] if rule_p && rule_p['rulespack_id']
          res[:test]         = rule_p['test']         if rule_p && rule_p['test']
          res[:infos]        = payload['infos']       if payload['infos']
          res[:time]         = attack.time
          res[:client_ip]    = request_p[:addr]       if request_p && request_p[:addr]
          res[:request]      = request_p              if request_p
          res[:params]       = payload['params']      if payload['params']
          res[:context]      = payload['context']     if payload['context']
          res[:headers]      = payload['headers']     if payload['headers']
          res
        end

        # @param [Sqreen::RequestRecord] rr
        def convert_request_record(rr)
          res = { :version => '20171208' }
          payload = rr.payload

          if payload[:observed]
            res[:observed] = payload[:observed].dup
            rulespack = nil
            if rr.observed[:attacks]
              res[:observed][:attacks] = rr.observed[:attacks].map do |att|
                natt = att.dup
                [:attack_type, :block].each { |k| natt.delete(k) } # signals stuff
                rulespack = natt.delete(:rulespack_id) || rulespack
                natt
              end
            end
            if rr.observed[:sqreen_exceptions]
              res[:observed][:sqreen_exceptions] = rr.observed[:sqreen_exceptions].map do |exc|
                nex = exc.dup
                excp = nex.delete(:exception)
                if excp
                  nex[:message] = excp.message
                  nex[:klass] = excp.class.name
                end
                rulespack = nex.delete(:rulespack_id) || rulespack
                nex
              end
            end
            res[:rulespack_id] = rulespack unless rulespack.nil?
            if rr.observed[:observations]
              res[:observed][:observations] = rr.observed[:observations].map do |cat, key, value, time|
                { :category => cat, :key => key, :value => value, :time => time }
              end
            end
            if rr.observed[:sdk] # rubocop:disable Style/IfUnlessModifier
              res[:observed][:sdk] = rr.processed_sdk_calls
            end
          end
          res[:local] = payload['local'] if payload['local']
          if payload['request']
            res[:request] = payload['request'].dup
            res[:client_ip] = res[:request].delete(:client_ip) if res[:request][:client_ip]
          else
            res[:request] = {}
          end
          if payload['response']
            res[:response] = payload['response'].dup
          else
            res[:response] = {}
          end

          res[:request][:parameters] = payload['params'] if payload['params']
          res[:request][:headers] = payload['headers'] if payload['headers']

          res = Sqreen::Kit::StringSanitizer.sanitize(res)

          if rr.redactor
            res[:request], redacted = rr.redactor.redact(res[:request])
            redacted = redacted.uniq
            if redacted.any? && res[:observed] && res[:observed][:attacks]
              res[:observed][:attacks] = WafRedactions.redact_attacks!(res[:observed][:attacks], redacted)
            end
            if redacted.any? && res[:observed] && res[:observed][:sqreen_exceptions]
              res[:observed][:sqreen_exceptions] = WafRedactions.redact_exceptions!(res[:observed][:sqreen_exceptions], redacted)
            end
          end

          res
        end

        # @param exception_evt [Sqreen::RemoteException]
        def convert_exception(exception_evt)
          payload = exception_evt.payload
          exception = payload['exception']
          ev = {
            :klass => exception.class.name,
            :message => exception.message,
            :params => payload['request_params'],
            :time => payload['time'],
            :infos => {
              :client_ip => payload['client_ip'],
            },
            :request => payload['request_infos'],
            :headers => payload['headers'],
            :rule_name => payload['rule_name'],
            :rulespack_id => payload['rulespack_id'],
          }

          ev[:infos].merge!(payload['infos']) if payload['infos']
          return ev unless exception.backtrace
          ev[:context] = { :backtrace => exception.backtrace.map(&:to_s) }
          ev
        end

        # @param [Sqreen::AggregatedMetric] agg_metric
        def convert_agg_metric(agg_metric)
          {
            name: agg_metric.name,
            observation: agg_metric.data,
            start: agg_metric.start,
            finish: agg_metric.finish,
          }
        end
      end
    end
  end
end
