# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'sqreen/aggregated_metric'
require 'sqreen/metrics'
require 'sqreen/mono_time'
require 'sqreen/metrics_store/unknown_metric'
require 'sqreen/metrics_store/unregistered_metric'
require 'sqreen/metrics_store/already_registered_metric'

module Sqreen
  # This store and register metrics
  class MetricsStore
    # definition keys
    NAME_KEY = 'name'.freeze
    KIND_KEY = 'kind'.freeze
    PERIOD_KEY = 'period'.freeze
    OPTIONS_KEY = 'options'.freeze

    # Currently ready samples
    attr_reader :store
    # All known metrics
    attr_reader :metrics

    def initialize
      @store = []
      @metrics = {} # name => (metric, period, start)
      @mutex = Mutex.new
    end

    # Definition contains a name,period and aggregate at least
    # @param definition [Hash] a metric definition
    # @param rule [RuleCB] the rule associated with this metric, if any
    # @param mklass [Object] Override metric object (used in testing)
    def create_metric(definition, rule = nil, mklass = nil)
      @mutex.lock

      name = definition[NAME_KEY]
      kind = definition[KIND_KEY]
      klass = valid_metric(kind, name)

      opts = definition[OPTIONS_KEY]
      metric = mklass || klass.new(opts)
      @metrics[name] = [
        metric,
        definition[PERIOD_KEY],
        nil # Start
      ]
      metric.name = name
      metric.rule = rule
      metric.period = definition[PERIOD_KEY]
      metric
    ensure
      @mutex.unlock
    end

    def metric?(name)
      @metrics.key?(name)
    end

    # @param at [Time] when is the store emptied
    def update(name, at, key, value)
      @mutex.lock
      metric, period, start = @metrics[name]
      raise UnregisteredMetric, "Unknown metric #{name}" unless metric
      next_sample(name, at) if start.nil? || (start + period) < at
      metric.update(key, value)
    ensure
      @mutex.unlock
    end

    # Drains every metrics and returns the store content
    # @param at [Time] when is the store emptied
    def publish(flush = true, at = Sqreen.time)
      @mutex.lock
      @metrics.each do |name, (_, period, start)|
        next_sample(name, at) if flush || !start.nil? && (start + period) < at
      end
      out = @store
      @store = []
      out
    ensure
      @mutex.unlock
    end

    protected

    def next_sample(name, at)
      metric = @metrics[name][0]
      r = metric.next_sample(at)
      @metrics[name][2] = at # new start
      return unless r

      r[NAME_KEY] = name
      obs = r[Metric::OBSERVATION_KEY]
      return unless obs && (!obs.respond_to?(:empty?) || !obs.empty?)
      start_of_mono = Time.now.utc - Sqreen.time

      agg = AggregatedMetric.new
      agg.metric = metric
      agg.rule = agg.metric.rule
      agg.start = start_of_mono + r[Metric::START_KEY]
      agg.finish = start_of_mono + r[Metric::FINISH_KEY]
      agg.data = obs
      @store << agg
    end

    def valid_metric(kind, name)
      unless Sqreen::Metric.const_defined?(kind)
        raise UnknownMetric, "No such #{kind} metric"
      end
      klass = Sqreen::Metric.const_get(kind)
      metric = @metrics[name] && @metrics[name][0]
      if metric && metric.class != klass
        msg = "Was a #{metric.class.name} not a #{klass.name} "
        raise AlreadyRegisteredMetric, msg
      end
      klass
    end
  end
end
