# typed: ignore

# Copyright (c) 2015 Sqreen. All Rights Reserved.
# Please refer to our terms for more information: https://www.sqreen.com/terms.html

require 'ipaddr'
require 'sqreen/node'

# TODO: move to Sqreen::IP

module Sqreen
  Trie = Struct.new(:head, :num_active_nodes, :family) do
    attr_reader :max_num_bits

    def initialize(*args)
      super
      self.family ||= Socket::AF_INET
      self.num_active_nodes = 0
      @max_num_bits = family == Socket::AF_INET ? 32 : 128
    end

    def insert(arg_prefix)
      raise 'family mismatch' if arg_prefix.family != family

      if head.nil?
        node = Node.new(
            arg_prefix.bitlen, # bit
            arg_prefix,
            nil, nil, nil, #l , r, parent
        )
        self.head = node
        self.num_active_nodes += 1
        return node
      end

      arg_addr = arg_prefix.address
      arg_bitlen = arg_prefix.bitlen
      xcur = self.head

      # descend until we find the end of the tree or go past the
      # bitlen of the prefix we're adding
      while xcur.bit < arg_bitlen || xcur.empty?
        if bit_set?(arg_addr, xcur.bit)
          break if xcur.r.nil?
          xcur = xcur.r
        else
          break if xcur.l.nil?
          xcur = xcur.l
        end
      end

      trie_addr = xcur.prefix.address

      # find first bit that differs between addr and trie_addr
      cmp_bit_end = [arg_bitlen, xcur.bit].min # after last to be compared

      differ_bit = (0...cmp_bit_end).find do |i|
        bit_set?(arg_addr, i) ^ bit_set?(trie_addr, i)
      end
      differ_bit = cmp_bit_end if differ_bit.nil?

      # go up till we find the parent before the bits differ
      xparent = xcur.parent
      while !xparent.nil? && xparent.bit >= differ_bit
        xcur = xparent
        xparent = xcur.parent
      end

      # case 1: replace current node's prefix
      if differ_bit == arg_bitlen && xcur.bit == arg_bitlen
        xcur.prefix = arg_prefix
        return xcur
      end

      xnew = Node.new(arg_prefix.bitlen, arg_prefix, nil, nil, nil)
      self.num_active_nodes += 1

      # case 2: append below found node
      if xcur.bit == differ_bit
        xnew.parent = xcur
        if bit_set?(arg_addr, xcur.bit)
          xcur.r = xnew
        else
          xcur.l = xnew
        end

        return xnew
      end

      # case 3: take place of found node
      if arg_bitlen == differ_bit
        if bit_set?(trie_addr, arg_bitlen)
          xnew.r = xcur
        else
          xnew.l = xcur
        end

        xnew.parent = xcur.parent

        if xcur.parent.nil?
          self.head = xnew
        elsif xcur.parent.r.equal?(xcur)
          xcur.parent.r = xnew
        else
          xcur.parent.l = xnew
        end

        xcur.parent = xnew

        return xnew
      end

      # case 4: need to add intermediate node to be parent of the
      #         both xnew and xcur

      # last arg is the parent of the new node
      xglue = Node.new(differ_bit, nil, nil, nil, xcur.parent)
      self.num_active_nodes += 1

      if bit_set?(arg_addr, differ_bit)
        xglue.r = xnew
        xglue.l = xcur
      else
        xglue.r = xcur
        xglue.l = xnew
      end

      xnew.parent = xglue

      if xcur.parent.nil?
        self.head = xglue
      elsif xcur.equal?(xcur.parent.r)
        xcur.parent.r = xglue
      else
        xcur.parent.l = xglue
      end

      xcur.parent = xglue

      xnew
    end

    # generate pdf with `dot -Tpdf <input> > foo.pdf`
    def to_dot(header = true)
      res = ''
      res = "graph trie {\n" if header

      next_name = 'a'
      obj_label_map = Hash.new do |h, k|
        h[k] = next_name
        next_name = next_name.next
        h[k]
      end

      head.walk(max_num_bits, true) do |node|
        node_name = obj_label_map[node.object_id]

        if node.empty?
          label = "<glue at bit #{node.bit}>"
        else
          ip_addr = IPAddr.new(node.prefix.address, node.prefix.family)
          label = ip_addr.to_s + '/' + node.prefix.bitlen.to_s
        end
        res += "#{node_name} [label=\"#{label}\"]\n"
        res += "#{node_name} -- #{obj_label_map[node.l.object_id]} [label=\"l\"]\n" if node.l
        res += "#{node_name} -- #{obj_label_map[node.r.object_id]} [label=\"r\"]\n" if node.r
      end
      res += "}\n" if header
      res
    end

    def search_best(addr, family)
      nodes = search_common(addr, family)
      for i in (nodes.size - 1).downto(0)
        xcur = nodes[i]
        return node_to_ip_addr(xcur) if xcur.prefix.matches?(addr, family)
      end

      nil
    end

    def search_matching(addr, family)
      nodes = search_common(addr, family)
      nodes.select { |xcur| xcur.prefix.matches?(addr, family) }
           .map { |xcur| node_to_ip_addr(xcur) }
           .reverse
    end

    private

    def node_to_ip_addr(node)
      ret = IPAddr.new(node.prefix.address, node.prefix.family)
      ret.send(:mask!, node.prefix.bitlen)
      ret.singleton_class.send(:define_method, :data) { node.prefix.data }
      ret
    end

    def bit_set?(addr, i)
      addr[max_num_bits - i - 1] == 1
    end

    def search_common(addr, family)
      raise 'family mismatch' unless family == self.family

      xstack = []

      xcur = head
      while !xcur.nil?
        unless xcur.empty?
          xstack << xcur
        end

        xcur = bit_set?(addr, xcur.bit) ? xcur.r : xcur.l
      end

      xstack
    end
  end
end
